import{j as e}from"./index-G4l7qcMm.js";import{B as i}from"./browse-Ypk-yEjt.js";import{F as l}from"./Stack-Cwu0vGSf.js";import{I as n,S as s,M as m,F as c}from"./Tooltip-DDkR3tcv.js";const u=r=>{const a=Array.from(Array(r.npartitions).keys()),o=t=>{r.setValue(t.target.value)};return e.jsx(i,{children:e.jsxs(l,{sx:{my:2},children:[e.jsx(n,{id:"partition-select-helper-label",children:"Partition"}),e.jsx(s,{labelId:"partition-select-label",id:"partition-select",value:r.value,label:"Partition",onChange:o,children:a.map(t=>e.jsx(m,{value:t,children:t},`partition-${t}`))}),e.jsx(c,{children:"A portion of the rows"})]})})};export{u as C};
