import{q as p,k as f,r as w,t as x,_ as F,b as e,j as d,o as C,c as G,v as y}from"./index-G4l7qcMm.js";import{b as R,g}from"./Input-BJwdycW3.js";function j(o){return p("MuiFormGroup",o)}f("MuiFormGroup",["root","row","error"]);const v=["className","row"],M=o=>{const{classes:r,row:t,error:s}=o;return y({root:["root",t&&"row",s&&"error"]},j,r)},U=C("div",{name:"MuiFormGroup",slot:"Root",overridesResolver:(o,r)=>{const{ownerState:t}=o;return[r.root,t.row&&r.row]}})(({ownerState:o})=>e({display:"flex",flexDirection:"column",flexWrap:"wrap"},o.row&&{flexDirection:"row"})),D=w.forwardRef(function(r,t){const s=x({props:r,name:"MuiFormGroup"}),{className:a,row:l=!1}=s,c=F(s,v),i=R(),u=g({props:s,muiFormControl:i,states:["error"]}),n=e({},s,{row:l,error:u.error}),m=M(n);return d.jsx(U,e({className:G(m.root,a),ownerState:n,ref:t},c))});export{D as F};
