const __vite__mapDeps=(i,m=__vite__mapDeps,d=(m.f||(m.f=["assets/array-1d-B1Aq4rcu.js","assets/index-G4l7qcMm.js","assets/index-Bjnup-cL.css","assets/browse-Ypk-yEjt.js","assets/index.esm-Dfo1uwtt.js","assets/Input-BJwdycW3.js","assets/Stack-Cwu0vGSf.js","assets/FormGroup-y9-iIeyA.js","assets/FormControlLabel-Ckn-i3Sp.js","assets/array-nd-BY-MA2EH.js"])))=>i.map(i=>d[i]);
import{j as t,J as n,r,S as a,I as s}from"./index-G4l7qcMm.js";const i=r.lazy(()=>s(()=>import("./array-1d-B1Aq4rcu.js"),__vite__mapDeps([0,1,2,3,4,5,6,7,8]))),c=r.lazy(()=>s(()=>import("./array-nd-BY-MA2EH.js"),__vite__mapDeps([9,1,2,3,4,5]))),h=e=>{if(e.structure.data_type.hasOwnProperty("fields"))return t.jsxs(n,{severity:"warning",children:['This is a "record array" with a'," ",t.jsx("a",{href:"https://numpy.org/doc/stable/user/basics.rec.html",target:"_blank",rel:"noreferrer",children:"structured data type"}),'. The web interface cannot view it. Use the "Download" tab to access the data.']});switch(e.structure.shape.length<2){case!0:return t.jsx(r.Suspense,{fallback:t.jsx(a,{variant:"rectangular"}),children:t.jsx(i,{segments:e.segments,item:e.item,link:e.link,structure:e.structure})});case!1:return t.jsx(r.Suspense,{fallback:t.jsx(a,{variant:"rectangular"}),children:t.jsx(c,{segments:e.segments,item:e.item,link:e.link,structure:e.structure})})}};export{h as default};
