import{j as e,w as a,S as r}from"./index-G4l7qcMm.js";import{B as i}from"./browse-Ypk-yEjt.js";import{h as n}from"./default-highlight-BScOjbTS.js";import"./toPropertyKey-DCwh5dYN.js";const m=t=>t.json!==void 0?e.jsxs(i,{children:[e.jsx(a,{id:"metadata-title",variant:"h6",component:"h2",children:"Detailed Machine-Readable Representation of Item"}),e.jsx(n,{language:"json",children:JSON.stringify(t.json,null,2)})]}):e.jsx(r,{variant:"rectangular"});export{m as default};
